#
# Copyright 2021 Canonical Ltd.  All rights reserved.
#

import os
import subprocess
import six
import time

from charmhelpers import fetch
from charmhelpers.core.hookenv import log

CURRENT_VERSION = '1.10'

# This contains a list of old versions. It's not the most efficient way
# to deal with the problem that we don't know what package are installed
# but lets us solve the problem for the time being without implementing
# more complex ways of determine installed packages
OLD_VERSIONS = ['1.9']


APT_NO_LOCK = 100  # The return code for "couldn't acquire lock" in APT.
CMD_RETRY_DELAY = 10  # Wait 10 seconds between command retries.
CMD_RETRY_COUNT = 3  # Retry a failing fatal command X times.


def _apt_pin_path(name):
    return "/etc/apt/preferences.d/{}".format(name)


def _remove_package_pin(name):
    """Remove a pin introduced by calling pin_package_version"""
    pin_path = _apt_pin_path(name)
    if not os.path.exists(pin_path):
        return

    os.remove(pin_path)


def _run_with_retries(cmd, max_retries=CMD_RETRY_COUNT, retry_exitcodes=(1,),
                      retry_message="", cmd_env=None, quiet=False):
    """Run a command and retry until success or max_retries is reached.
    :param cmd: The apt command to run.
    :type cmd: str
    :param max_retries: The number of retries to attempt on a fatal
                        command. Defaults to CMD_RETRY_COUNT.
    :type max_retries: int
    :param retry_exitcodes: Optional additional exit codes to retry.
                            Defaults to retry on exit code 1.
    :type retry_exitcodes: tuple
    :param retry_message: Optional log prefix emitted during retries.
    :type retry_message: str
    :param: cmd_env: Environment variables to add to the command run.
    :type cmd_env: Option[None, Dict[str, str]]
    :param quiet: if True, silence the output of the command from stdout and
        stderr
    :type quiet: bool
    """
    env = fetch.get_apt_dpkg_env()
    if cmd_env:
        env.update(cmd_env)

    kwargs = {}
    if quiet:
        devnull = os.devnull if six.PY2 else subprocess.DEVNULL
        kwargs['stdout'] = devnull
        kwargs['stderr'] = devnull

    if not retry_message:
        retry_message = "Failed executing '{}'".format(" ".join(cmd))
    retry_message += ". Will retry in {} seconds".format(CMD_RETRY_DELAY)

    retry_count = 0
    result = None

    retry_results = (None,) + retry_exitcodes
    while result in retry_results:
        try:
            result = subprocess.check_call(cmd, env=env, **kwargs)
        except subprocess.CalledProcessError as e:
            retry_count = retry_count + 1
            if retry_count > max_retries:
                raise
            result = e.returncode
            log(retry_message)
            time.sleep(CMD_RETRY_DELAY)


def _run_apt_command(cmd, fatal=False, quiet=False):
    """Run an apt command with optional retries.
    :param cmd: The apt command to run.
    :type cmd: str
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    :param quiet: if True, silence the output of the command from stdout and
        stderr
    :type quiet: bool
    """
    if fatal:
        _run_with_retries(
            cmd, retry_exitcodes=(1, APT_NO_LOCK,),
            retry_message="Couldn't acquire DPKG lock",
            quiet=quiet)
    else:
        kwargs = {}
        if quiet:
            devnull = os.devnull if six.PY2 else subprocess.DEVNULL
            kwargs['stdout'] = devnull
            kwargs['stderr'] = devnull
        subprocess.call(cmd, env=fetch.get_apt_dpkg_env(), **kwargs)


def _format_package_version(version):
    return version.replace(".", "")


def _versioned_package_name(name, version=None):
    if not version:
        version = CURRENT_VERSION
    return "{}-{}".format(name, _format_package_version(version))


def _is_package_installed(name):
    try:
        subprocess.check_call(['/usr/bin/dpkg', '-s', name])
        return True
    except subprocess.CalledProcessError as ex:
        return False


def _remove_old_versions_for_package(name):
    for version in OLD_VERSIONS:
        # As 1.9 doesn't have versioned package names we need to add special
        # handling for it
        if version == '1.9':
            pkg_name = name
        else:
            pkg_name = _versioned_package_name(name, version)

        if not _is_package_installed(pkg_name):
            continue

        # The charmhelpers package doesn't provide support to just remove
        # a package but not purging it which we can't do here as that would
        # also remove the snap which we don't want to remove
        _run_apt_command(['apt-get', 'remove', '--assume-yes', pkg_name], fatal=True)


def install_or_upgrade_anbox_package(name):
    # Remove APT pin previous versions of the charm have put in place
    _remove_package_pin(name)

    fetch.apt_update(fatal=True)

    # First we have to get rid of the old package to avoid file conflicts. As
    # this does a remove and not a purge the snap will remain installed and will
    # correctly updated with the next step
    _remove_old_versions_for_package(name)

    # Second we install the latest versioned package from the archive
    pkg_name = _versioned_package_name(name)
    fetch.apt_install([pkg_name], fatal=True)


def remove_anbox_package(name):
    # Remove APT pin previous versions of the charm have put in place
    _remove_package_pin(name)

    pkg_name = _versioned_package_name(name)

    # Don't attempt to purge package if it is not installed
    if not _is_package_installed(pkg_name):
        return

    fetch.apt_purge([pkg_name], fatal=True)
