#
# Anbox - The Android in a Box runtime environment
# Copyright 2020 Canonical Ltd.  All rights reserved.
#

import json
import os
import re
import urllib.request
from subprocess import check_call, check_output, CalledProcessError

from charmhelpers import fetch
from charmhelpers.core.hookenv import local_unit, log

UA_STATUS_PATH = "/var/lib/ubuntu-advantage/status.json"
UA_MACHINE_TOKEN_PATH = "/var/lib/ubuntu-advantage/private/machine-token.json"

CONTRACTS_SERVICE_STAGING_URL = "https://contracts.staging.canonical.com"

UA_CLIENT_PREMIUM_PPA = "ppa:canonical-server/ua-client-premium"

unit_name = local_unit().replace("/", "-")
ANBOX_CLOUD_APT_SOURCES_FILE = "/etc/apt/sources.list.d/{}-anbox-cloud.list".format(unit_name)
ANBOX_CLOUD_APT_AUTH_FILE = "/etc/apt/auth.conf.d/90{}-anbox-cloud".format(unit_name)
ANBOX_CLOUD_APT_KEY_FILE = '/etc/apt/trusted.gpg.d/{}-anbox-cloud.asc'.format(unit_name)

UBUNTU_KEYSERVER_URL = 'https://keyserver.ubuntu.com'

ANBOX_CLOUD_STABLE_ARCHIVE_KEY_ID = "0x6B787E5FB493FA4FD0095D106DEC2758ACD5A973"

"""
Anbox Cloud UA Layer

Grants access to Anbox Cloud private archive using either a UA token, or
custom source and key.

The custom source will always take precedence over the UA token, however the token
will always be attached if available.
If ua_token is removed, a `ua detach` is performed.
If both the ua_token and the custom sources are removed, the archives are removed as well.
Changing `ua_token` to an invalid one will perform a detach on the machine but the
sources will remain configured.

The `ua.ready` flag is set as long as at least one access method is configured.
The `ua.changed` flag is set when the archive has changed. APT will be updated by that
point so there's no need to do it again .

On stop, a detach is performed, and the archives are removed if this charm is the one
that attached the token
"""


def require_ua_client(func):
    """
    Decorator used to ensure the UA client is present, and installs it if not
    """
    def function_wrapper(*args, **kwargs):
        version = 0
        try:
            raw_version = str(check_output(['ua', 'version']), 'utf-8')
            log('UA client version is {}'.format(raw_version), level='INFO')
            version = int(raw_version.split('.')[0])
        except Exception as err:
            log('Failed to determine UA client version: {}'.format(err), level='ERROR')
            raise

        if version >= 20:
            func(*args, **kwargs)
            return

        fetch.apt_install(['software-properties-common'], fatal=True)

        try:
            check_call(['add-apt-repository', UA_CLIENT_PREMIUM_PPA])
        except CalledProcessError as err:
            raise Exception("Failed to setup UA PPA: {}".format(err))

        fetch.apt_update(fatal=False)
        fetch.apt_install(['ubuntu-advantage-tools'], fatal=True)

        func(*args, **kwargs)
    return function_wrapper


def set_custom_source(ua_source, ua_source_key):
    if len(ua_source) == 0 and len(ua_source_key) > 0:
        raise ValueError('ua_source must be set if ua_source_key is not empty')
    elif len(ua_source) > 0 and len(ua_source_key) == 0:
        raise ValueError('ua_source_key must be set if ua_source is not empty')

    # Sanitize inputs if they are read from a file
    ua_source = ua_source.strip()
    ua_source_key = ua_source_key.strip()

    ppa_info = ua_source.split()
    if len(ppa_info) < 4:
        raise ValueError("malformed ua_source")

    ppa_url = ppa_info[1]
    ppa_series = ppa_info[2]

    ensure_series_is_supported(get_current_series(), ppa_series)

    # Remove credentials from url and store them securely elsewhere
    creds = re.search('(?<=//).*(?=@)', ppa_url)
    if not creds or not creds.group(0):
        raise ValueError('invalid ua_source: missing authentication')

    creds = creds.group(0)
    safe_url = ppa_url.replace('{}@'.format(creds), '')

    creds = creds.split(':')
    if len(creds) != 2:
        raise ValueError('invalid ua_source: malformed authentication')
    username = creds[0]
    password = creds[1]

    public_key = None
    if ua_source_key == ANBOX_CLOUD_STABLE_ARCHIVE_KEY_ID:
        log('Taking stable archive key from the charm for {}'.format(ua_source_key), level='DEBUG')
        with open('files/anbox-cloud.pem') as f:
            public_key = f.read()
    else:
        if not ua_source_key.startswith('0x'):
            ua_source_key = '0x{}'.format(ua_source_key)
        search_url = '{}/pks/lookup?op=get&search={}'.format(UBUNTU_KEYSERVER_URL, ua_source_key)
        log('Fetching key with fingerprint {} at {}'.format(ua_source_key, search_url), level='DEBUG')
        public_key = urllib.request.urlopen(search_url, timeout=300).read().decode("utf-8")

    add_ppa(safe_url, username, password, public_key)


def set_ua_token(ua_token: str, use_staging: bool = False):
    try:
        try_ua_attach(ua_token, use_staging)
    except Exception as e:
        log('Failed to attach: {}'.format(e), level='ERROR')
        raise

    # ua_token was unset
    if not os.path.exists(UA_MACHINE_TOKEN_PATH):
        return

    with open(UA_MACHINE_TOKEN_PATH) as machine_token_file:
        data = json.load(machine_token_file)
        has_anbox_cloud = False
        for resource in data['availableResources']:
            if resource['name'] == "anbox-cloud" and resource['available'] is True:
                has_anbox_cloud = True
                break

        if not has_anbox_cloud:
            raise Exception("Not entitled for Anbox Cloud")

        contract_info = data['machineTokenInfo']['contractInfo']
        ent_by_name = dict((e["type"], e) for e in contract_info["resourceEntitlements"])

        ent = ent_by_name['anbox-cloud']

        supported_series = ent['affordances']['series']
        current_series = get_current_series()
        ensure_series_is_supported(current_series, supported_series)

        supported_architectures = ent['affordances']['architectures']
        ensure_architecture_is_supported(supported_architectures)

        tokens_by_name = dict((ent["type"], ent["token"]) for ent in data.get("resourceTokens", []))
        apt_resource_token = tokens_by_name['anbox-cloud']
        apt_url = ent['directives']['aptURL']

        public_key = None
        with open('files/anbox-cloud.pem') as f:
            public_key = f.read()
        if public_key is None:
            raise Exception('Failed to read archive public key')

        add_ppa(apt_url, 'bearer', apt_resource_token, public_key)


@require_ua_client
def try_ua_attach(ua_token: str, use_staging: bool = False):
    # We can't attach if we are already attached
    try_ua_detach()

    # Sanitize inputs if they are read from a file
    ua_token = ua_token.strip()

    args = ['ua', 'attach', ua_token]
    log('Attaching UA token (staging: {})'.format(use_staging), level='INFO')
    if use_staging:
        args = ['env', 'UA_CONTRACT_URL={}'.format(CONTRACTS_SERVICE_STAGING_URL)] + args

    check_call(args)


@require_ua_client
def try_ua_detach():
    """Attempt to detach the UA token if already attached. Otherwise does nothing"""
    attached = True
    if not os.path.exists(UA_STATUS_PATH):
        attached = False
    else:
        with open(UA_STATUS_PATH, 'r') as f:
            status = json.loads(f.read())
            attached = status['attached']

    if not attached:
        log('UA not attached, not attempting to detach', level='INFO')
        return

    was_attached_with_staging = False
    with open(ANBOX_CLOUD_APT_SOURCES_FILE, 'r') as f:
        current_source = f.read()
        # Hack to determine if the previous attachment was done in a staging environment
        if 'staging.anbox-cloud.io' in current_source:
            was_attached_with_staging = True

    log('Detaching UA token (staging: {})'.format(was_attached_with_staging), level='INFO')
    args = ['ua', 'detach', '--assume-yes']
    if was_attached_with_staging:
        args = ['env', 'UA_CONTRACT_URL={}'.format(CONTRACTS_SERVICE_STAGING_URL)] + args
    check_call(args)


def add_ppa(url: str, login: str, password: str, ppa_key: str):
    """
    Add a PPA to the system sources and updates apt cache. Overrides previous
    PPA if it exists
    :param: url (str) URL to the PPA. It should not contain the authentication token
    :param: login (str) PPA username
    :param: password (str) PPA password
    :param: ppa_key (str) PPA public key
    """
    need_apt_update = False

    with open(ANBOX_CLOUD_APT_SOURCES_FILE, 'w+') as f:
        source = "deb {} {} main\n".format(url, get_current_series())
        if f.read() != source:
            f.seek(0)
            f.write(source)
            need_apt_update = True

    # APT on bionic can't deal with URLs in /etc/apt/auth.conf.d which have a protocol set
    # See apt_auth.conf(5)
    if login and password:
        auth_url = url.replace('https://', '')
        with open(os.open(ANBOX_CLOUD_APT_AUTH_FILE, os.O_CREAT | os.O_RDWR, 0o600), 'w+') as f:
            auth = "machine {} login {} password {} # Anbox Cloud".format(auth_url, login, password)
            if f.read() != auth:
                f.seek(0)
                f.write(auth)
                need_apt_update = True

    if ppa_key:
        with open(os.open(ANBOX_CLOUD_APT_KEY_FILE, os.O_CREAT | os.O_RDWR, 0o644), 'w+') as f:
            if f.read() != ppa_key:
                f.seek(0)
                f.write(ppa_key)
                need_apt_update = True

    if need_apt_update:
        fetch.apt_update(fatal=True)


def get_current_series():
    return str(check_output(['lsb_release', '-c', '-s']), 'utf-8').strip('\n')


def ensure_series_is_supported(current_series, supported_series):
    if current_series not in supported_series:
        raise Exception("Current Ubuntu series {} is unsupported".format(current_series))


def ensure_architecture_is_supported(supported_architectures):
    current_arch = str(check_output(['uname', '-m']), 'utf-8').strip('\n')
    if current_arch == "aarch64":
        current_arch = "arm64"
    if current_arch not in supported_architectures:
        raise Exception("Current architecture {} is unsupported".format(current_arch))