#!/usr/bin/env python3
#
# (c) 2020 Canonical Ltd. All right reserved
#

import logging
import os

from ops.charm import CharmEvents, CharmBase, RelationChangedEvent
from ops.framework import Object, EventBase, EventSource, StoredState

logger = logging.getLogger(__name__)


class AmsChangedEvent(EventBase):
    pass


class AmsEvents(CharmEvents):
    ams_changed = EventSource(AmsChangedEvent)


class AmsClient(Object):
    on = AmsEvents()
    state = StoredState()

    def __init__(self, charm: CharmBase, relation_name: str, cert_path: str):
        super().__init__(charm, relation_name)
        self._relation_name = relation_name

        self.state.set_default(api_address=None)
        self.state.cert_path = cert_path

        self.framework.observe(charm.on[relation_name].relation_joined, self.on_relation_joined)
        self.framework.observe(charm.on[relation_name].relation_changed, self.on_relation_changed)
        self.framework.observe(charm.on[relation_name].relation_departed, self.on_relation_departed)


    def on_relation_joined(self, event):
        if not os.path.exists(self.state.cert_path):
            logger.warning("deferring configuration due to missing certificate")
            event.defer()
            return

        remote_data = self.framework.model.get_relation(self._relation_name).data[self.model.unit]
        with open(self.state.cert_path) as cert:
            remote_data['client_certificate'] = cert.read()
        logger.info("configuration done")


    def on_relation_changed(self, event: RelationChangedEvent):
        # FIXME: gracefully handle HA AMS
        remote_data = event.relation.data[next(iter(event.relation.units))]
        addr = remote_data.get('private_address', None)
        port = remote_data.get('port', None)
        if addr is None or port is None:
            # AMS didn't send its information yet
            return

        self.state.api_address = '{}:{}'.format(addr.replace('"', ''), port)
        self.on.ams_changed.emit()


    def on_relation_departed(self, event):
        self.state.api_address = None


    @property
    def is_joined(self):
        return self.framework.model.get_relation(self._relation_name) is not None


    @property
    def is_ready(self) -> bool:
        return self.state.api_address is not None


    @property
    def api_address(self):
        return self.state.api_address
