#!/usr/bin/env python3
#
# (c) 2020 Canonical Ltd. All right reserved
#

import json
import logging
from typing import Mapping

from ops.charm import CharmEvents, CharmBase, RelationChangedEvent, RelationJoinedEvent
from ops.framework import Object, EventBase, EventSource, StoredState

logger = logging.getLogger(__name__)


class Certificate(dict):
    def __init__(self, cert_type: str, common_name: str, cert: str, key: str):
        super().__init__({
            'cert_type': cert_type,
            'common_name': common_name,
            'cert': cert,
            'key': key
        })

    @property
    def cert_type(self):
        return self['cert_type']

    @property
    def common_name(self):
        return self['common_name']

    @property
    def cert(self):
        return self['cert']

    @property
    def key(self):
        return self['key']

    def __hash__(self):
        return hash((self.cert_type, self.cert))


class ClientCertAvailable(EventBase):
    pass


class ServerCertAvailable(EventBase):
    pass


class RootCACertAvailable(EventBase):
    pass


class CertificateEvents(CharmEvents):
    ca_cert_available = EventSource(RootCACertAvailable)
    client_cert_available = EventSource(ClientCertAvailable)
    server_cert_available = EventSource(ServerCertAvailable)


class Certificates(Object):
    on = CertificateEvents()
    state = StoredState()

    def __init__(self, charm: CharmBase, relation_name: str = 'certificates'):
        super().__init__(charm, relation_name)
        self._relation_name = relation_name
        self._unit_name = charm.model.unit.name.replace('/', '_')

        self.state.set_default(
            ca=None,
            server_certs=set(),
            client_certs=set()
        )

        self.framework.observe(charm.on[relation_name].relation_joined, self.on_relation_joined)
        self.framework.observe(charm.on[relation_name].relation_changed, self.on_relation_changed)
        # self.framework.observe(charm.on[relation_name].relation_departed, self.on_relation_departed)


    def on_relation_joined(self, event: RelationJoinedEvent):
        remote_data = event.relation.data[self.model.unit]
        remote_data['unit_name'] = self._unit_name


    def on_relation_changed(self, event: RelationChangedEvent):
        # CA
        ca_cert = self.ca_cert
        if ca_cert and self.state.ca != hash(ca_cert):
            self.state.ca = hash(ca_cert)
            self.on.ca_cert_available.emit()

        # Client certificates
        client_certs = self.client_certs
        client_cert_updated = False
        # We can't store the full certificate in local state, so we only store the
        # hash to differentiate new certs from old ones
        for cert in client_certs:
            if hash(cert) not in self.state.client_certs:
                self.state.client_certs.add(hash(cert))
                client_cert_updated = True
        if client_cert_updated:
            self.on.client_cert_available.emit()

        # Server certs
        server_certs = self.server_certs
        server_cert_updated = False
        for cert in server_certs:
            if hash(cert) not in self.state.server_certs:
                self.state.server_certs.add(hash(cert))
                server_cert_updated = True
        if server_cert_updated:
            self.on.server_cert_available.emit()


    def request_server_cert(self, common_name, sans=None):
        """
        Request a server certificate and key be generated for the given
        common name (`cn`) and optional list of alternative names (`sans`).

        This can be called multiple times to request more than one server
        certificate, although the common names must be unique. If called
        again with the same common name, it will be ignored.
        """
        relation = self.framework.model.get_relation(self._relation_name)
        data = relation.data[self.model.unit]

        requested_certs = data.get('cert_requests', {})
        requested_certs[common_name] = {'sans': sans or []}
        data['cert_requests'] = json.dumps(requested_certs)


    def request_client_cert(self, common_name, sans):
        """
        Request a client certificate and key be generated for the given
        common name (`cn`) and list of alternative names (`sans`).

        This can be called multiple times to request more than one client
        certificate, although the common names must be unique.  If called
        again with the same common name, it will be ignored.
        """
        relation = self.framework.model.get_relation(self._relation_name)
        data = relation.data[self.model.unit]

        requested_certs = data.get('client_cert_requests', {})
        requested_certs[common_name] = {'sans': sans}
        data['client_cert_requests'] = json.dumps(requested_certs)


    @property
    def ca_cert(self) -> Certificate:
        relation = self.framework.model.get_relation(self._relation_name)
        if not relation:
            return None

        for unit in relation.units:
            if unit.name == self._unit_name:
                continue

            received = relation.data[unit]
            ca_cert = received.get('ca', None)
            if ca_cert:
                return Certificate('ca', unit.name, ca_cert, None)
            return None


    @property
    def server_certs(self) -> Mapping[str, Certificate]:
        """
        Mapping of server [Certificate][] instances by their `common_name`.
        """
        relation = self.framework.model.get_relation(self._relation_name)
        if not relation:
            return None

        server_certs = []
        for unit in relation.units:
            if unit.name == self._unit_name:
                continue
            received = relation.data[unit]
            server_cert_field = '{}.processed_requests'.format(self._unit_name)
            server_certs_raw = json.loads(received.get(server_cert_field, '{}'))
            server_certs.extend([Certificate('server', common_name, cert['cert'], cert['key'])
                                 for common_name, cert in server_certs_raw.items()])
        return {cert.common_name: cert for cert in server_certs}


    @property
    def client_certs(self)  -> Mapping[str, Certificate]:
        """
        Mapping of client [Certificate][] instances by their `common_name`.
        """
        relation = self.framework.model.get_relation(self._relation_name)
        if not relation:
            return None

        client_certs = []
        for unit in relation.units:
            if unit.name == self._unit_name:
                continue
            received = relation.data[unit]
            client_cert_field = '{}.processed_client_requests'.format(self._unit_name)
            client_certs_raw = json.loads(received.get(client_cert_field, '{}'))
            client_certs.extend([Certificate('client', common_name, cert['cert'], cert['key'])
                                 for common_name, cert in client_certs_raw.items()])
        return {cert.common_name: cert for cert in client_certs}


    @property
    def is_joined(self):
        return self.framework.model.get_relation(self._relation_name) is not None
