#!/usr/bin/env python3
#
# (c) 2020 Canonical Ltd. All right reservered
#

import logging
import os
import random
import socket
import string
import sys
import subprocess

from pathlib import Path

sys.path.append('lib')
import yaml
import ops

from ams import AmsClient, AmsChangedEvent
from gateway import StreamGatewayClient, StreamGatewayChangedEvent
from certificates import Certificates
from http_rel import Http, HttpChangedEvent, HttpDepartedEvent
from pki import create_cert_files_if_needed
from charmhelpers.core.hookenv import open_port, unit_public_ip, unit_private_ip
from charmhelpers.core.host import service_restart, service
from ops.charm import CharmBase, CharmEvents, RelationJoinedEvent
from ops.framework import StoredState
from ops.main import main
from ops.model import (ModelError, BlockedStatus, ActiveStatus)
from ua.client import set_ua_token, set_custom_source
from anbox import install_or_upgrade_anbox_package
from nrpe.client import NRPEClient


DEB_PACKAGE_NAME = "anbox-cloud-dashboard"
SNAP_NAME = DEB_PACKAGE_NAME
SERVICE = "snap.anbox-cloud-dashboard.web.service"

DASHBOARD_COMMON = "/var/snap/anbox-cloud-dashboard/common"
SERVICE_DIR = "{}/service".format(DASHBOARD_COMMON)
TLS_DIR = "{}/tls".format(DASHBOARD_COMMON)

DASHBOARD_CONFIG = '{}/config.yaml'.format(SERVICE_DIR)
AUTH_DATABASE = '{}/auth.sqlite'.format(SERVICE_DIR)
AMS_CERT_PATH = '{}/ams.crt'.format(SERVICE_DIR)
AMS_KEY_PATH = '{}/ams.key'.format(SERVICE_DIR)
API_CERT_PATH = '{}/app.crt'.format(TLS_DIR)
API_KEY_PATH = '{}/app.key'.format(TLS_DIR)

SECRET_LENGTH = 64

logger = logging.getLogger(__name__)

StatusBlockedSnapNotInstalled = BlockedStatus("UA token missing")
StatusBlockedCertificatesMissing = BlockedStatus("Waiting for CA and certificates")


class AnboxCloudDashboardCharm(CharmBase):
    on = CharmEvents()
    state = StoredState()

    def __init__(self, *args):
        super().__init__(*args)

        self.state.set_default(
            snap_installed=False,
            snap_connected=False,
            jwt_secret=self._generate_secret(SECRET_LENGTH),
            app_secret=self._generate_secret(SECRET_LENGTH),
            config=None,
        )

        self.ams = AmsClient(self, 'ams', AMS_CERT_PATH)
        self.framework.observe(self.ams.on.ams_changed, self.on_ams_changed)

        self.gateway = StreamGatewayClient(self, 'gateway')
        self.framework.observe(self.gateway.on.gateway_changed, self.on_gateway_changed)

        self.certificates = Certificates(self, 'certificates')
        self.framework.observe(self.on['certificates'].relation_joined, self.on_certificates_relation_joined)
        self.framework.observe(self.certificates.on.server_cert_available, self.on_server_cert_available)

        self.nrpe = NRPEClient(self, 'nrpe-external-master')
        self.framework.observe(self.nrpe.on.nrpe_available, self.on_nrpe_available)

        self.reverseproxy = Http(self, 'reverseproxy')
        self.framework.observe(self.on['reverseproxy'].relation_joined, self.on_reverseproxy_available)
        self.framework.observe(self.reverseproxy.on.http_changed, self.on_reverseproxy_changed)
        self.framework.observe(self.reverseproxy.on.http_departed, self.on_reverseproxy_departed)

        self.framework.observe(self.on.install, self.on_install)
        self.framework.observe(self.on.stop, self.on_stop)
        self.framework.observe(self.on.config_changed, self.on_config_changed)
        self.framework.observe(self.on.upgrade_charm, self.on_upgrade_charm)

    # AMS related methods
    def on_ams_changed(self, event: AmsChangedEvent):
        if not self.state.snap_installed:
            event.defer()
            return

        self.configure()

    # Gateway related methods
    def on_gateway_changed(self, event: StreamGatewayChangedEvent):
        if not self.state.snap_installed:
            event.defer()
            return

        self.configure()

    # Reverse proxy related methods
    def on_reverseproxy_changed(self, event: HttpChangedEvent):
        self.configure()

    def on_reverseproxy_departed(self, event: HttpDepartedEvent):
        self.configure()

    # Certificate related methods
    def on_certificates_relation_joined(self, event: RelationJoinedEvent):
        common_name = unit_private_ip()
        subject_alt_names = list()
        subject_alt_names.append(socket.gethostname())
        subject_alt_names.append(unit_private_ip())
        self.certificates.request_server_cert(common_name, subject_alt_names)

    def on_server_cert_available(self, event):
        if not self.state.snap_installed:
            event.defer()
            return

        cert = self.certificates.server_certs.get(unit_private_ip(), None)
        if not cert:
            return

        with open(os.open(API_CERT_PATH, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
            f.write(cert.cert)

        with open(os.open(API_KEY_PATH, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as f:
            f.write(cert.key)
        self.configure()

    # NRPE related methods
    def on_nrpe_available(self, event):
        check_name = "check_{}".format(self.model.unit.name.replace("/", "_"))

        self.nrpe.add_check(command=[
            '/usr/lib/nagios/plugins/check_tcp',
            '-H', unit_private_ip(),
            '-p', str(self.model.config['port'] or 5000),
        ], name=check_name)

        # Save all new checks to filesystem and to Nagios
        self.nrpe.commit()

    # Reverse Proxy related methods
    def on_reverseproxy_available(self, event: RelationJoinedEvent):
        self.reverseproxy.configure(
            port=self.model.config['port'],
            service_name='app-anbox-cloud-dashboard'
        )

    def extract_snap_version(self):
        try:
            info = subprocess.run(['snap', 'info', SNAP_NAME],
                                stdout=subprocess.PIPE)
        except subprocess.CalledProcessError as err:
            pass
        else:
            stdout = info.stdout.decode('utf8').splitlines()
            version = [line.split()[1] for line in stdout if 'installed' in line]
            if version and len(version) == 1:
                return version[0]
        return None

    def set_charm_version(self):
        snap_version = self.extract_snap_version()
        if snap_version:
            self.model.unit.set_workload_version(snap_version)

    # Charm related methods
    def on_install(self, event: ops.charm.InstallEvent):
        self._try_install_snap()
        if not self.state.snap_installed:
            return

        create_cert_files_if_needed(AMS_CERT_PATH, AMS_KEY_PATH)
        self.configure()

    def on_stop(self, event: ops.charm.StopEvent):
        remove_anbox_package(DEB_PACKAGE_NAME)
        self.state.snap_connected = False

    def on_upgrade_charm(self, event: ops.charm.UpgradeCharmEvent):
        self._try_install_snap()
        if not self.state.snap_installed:
            event.defer()
            return

        create_cert_files_if_needed(AMS_CERT_PATH, AMS_KEY_PATH)
        self.configure()

    def on_config_changed(self, event: ops.charm.ConfigChangedEvent):
        self._try_install_snap()
        if not self.state.snap_installed:
            event.defer()
            return

        self.configure()

    @staticmethod
    def _generate_secret(length: int) -> str:
        letters = string.ascii_letters + string.digits
        return ''.join(random.choice(letters) for _ in range(length))

    def _try_install_snap(self):
        """
        _try_install_snap checks for all possible sources in order of priority for
        a snap to install:
        Resource > Custom PPA source > UA token
        If a source with higher precedence becomes available, it is used to install the
        snap.
        :return:
        """
        def finalize():
            if not os.path.exists(SERVICE_DIR):
                os.mkdir(SERVICE_DIR, mode=0o700)
                os.mkdir(TLS_DIR, mode=0o700)
                self.state.snap_installed = True

        if self.state.snap_installed:
            install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
            self.set_charm_version()
            return

        dashboard_snap = None
        try:
            dashboard_snap = self.model.resources.fetch('anbox-cloud-dashboard')
        except ModelError:
            logger.debug("no snap found in resource, trying UA")
        else:
            if dashboard_snap is not None and Path(dashboard_snap).stat().st_size > 0:
                try:
                    subprocess.check_output(['snap', 'install', '--dangerous', dashboard_snap])
                except OSError as e:
                    raise RuntimeError("failed to install snap package: {}".format(e))
                finalize()
                return

        ppa_source = self.model.config['ua_source']
        ppa_source_key = self.model.config['ua_source_key']
        ua_token = self.model.config['ua_token']
        use_staging = self.model.config['ua_use_staging']
        if ppa_source and ppa_source_key:
            set_custom_source(ppa_source, ppa_source_key)
            install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
        elif ua_token:
            set_ua_token(ua_token, use_staging)
            install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
        else:
            self.model.unit.status = StatusBlockedSnapNotInstalled
            return

        finalize()
        self.set_charm_version()


    def configure(self):
        if not self.state.snap_installed:
            self.model.unit.status = StatusBlockedSnapNotInstalled
            return
        if not os.path.exists(API_KEY_PATH) or not os.path.exists(API_CERT_PATH):
            self.model.unit.status = StatusBlockedCertificatesMissing
            return

        port = self.model.config['port'] or 5000
        open_port(port, 'TCP')

        private_address = unit_private_ip()

        # FIXME this is a workaround for the appliance when the charms > 1.9.0 are used with
        # the == 1.9.0 snap which has no configure hook which will let the command below fail
        snap_version = self.extract_snap_version()
        if snap_version != "1.9.0":
            subprocess.check_call(['snap', 'set', SNAP_NAME, 'listen-address={}:{}'.format(private_address, port)])

        # If a location is set it takes priority over the units address and the
        # configured port
        location = self.model.config['location']
        if location is None or len(location) == 0:
            location = self.reverseproxy.remote_public_address
        if location is None or len(location) == 0:
            dashboard_url = 'https://{}:{}'.format(unit_public_ip(), port)
        else:
            dashboard_url = 'https://{}'.format(location)

        cfg = {
            'AUTH_DATABASE_PATH': AUTH_DATABASE,
            'BASE_URI': dashboard_url,
            'JWT_SECRET': self.state.jwt_secret,
            'SECRET_KEY': self.state.app_secret,
            # The dashboard crashes without all keys, even if they are empty
            'AMS_API_URL': '',
            'AMS_API_CERTIFICATE': '',
            'AMS_API_CERTIFICATE_KEY': '',
            'ASG_API_URL': '',
            'ASG_API_TOKEN': '',
            # We set a default value here to not let the dashboard crash when it's
            # not set in the configuration.
            'VARIANT': 'regular',
        }

        variant = self.model.config['variant']
        if variant is not None and (variant == 'regular' or variant == 'appliance'):
            cfg['VARIANT'] = variant

        if self.ams.is_ready:
            cfg['AMS_API_URL'] = 'https://{}'.format(self.ams.api_address)
            cfg['AMS_API_CERTIFICATE'] = AMS_CERT_PATH
            cfg['AMS_API_CERTIFICATE_KEY'] = AMS_KEY_PATH

        if self.gateway.is_ready:
            cfg['ASG_API_URL'] = self.gateway.private_api_address
            cfg['ASG_API_TOKEN'] = self.gateway.client_token

        if cfg == self.state.config:
            logger.info("no changes when reconfiguring the dashboard")
            return
        self.state.config = cfg
        with open(os.open(DASHBOARD_CONFIG, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as config:
            config.truncate(0)
            config.write(yaml.safe_dump(cfg))

        if not self.state.snap_connected:
            subprocess.check_output(['snap', 'connect', 'anbox-cloud-dashboard:daemon-notify', 'system:daemon-notify'])
            service('enable', SERVICE)
            self.state.snap_connected = True

        service_restart(SERVICE)
        self.model.unit.status = ActiveStatus('')


if __name__ == "__main__":
    main(AnboxCloudDashboardCharm)
