#!/usr/bin/env python3
#
# (c) 2020 Canonical Ltd. All right reservered
#
import logging

from ops.charm import CharmEvents, RelationChangedEvent
from ops.framework import Object, EventBase, EventSource, StoredState

logger = logging.getLogger(__name__)


class StreamGatewayChangedEvent(EventBase):
    pass


class StreamGatewayClientEvents(CharmEvents):
    gateway_changed = EventSource(StreamGatewayChangedEvent)


class StreamGatewayClient(Object):
    on = StreamGatewayClientEvents()
    state = StoredState()

    def __init__(self, charm, relation_name):
        super().__init__(charm, relation_name)
        self._relation_name = relation_name

        self.state.set_default(
            api_address=None,
            private_api_address=None,
            client_token=None,
            allow_insecure=None,
        )

        self.framework.observe(charm.on[relation_name].relation_changed, self.on_relation_changed)
        self.framework.observe(charm.on[relation_name].relation_departed, self.on_relation_departed)

    def on_relation_changed(self, event: RelationChangedEvent):
        # FIXME: gracefully handle HA Stream Gateway
        remote_data = event.relation.data[next(iter(event.relation.units))]
        token = remote_data.get('api-token')
        api_address = remote_data.get('url')
        private_api_address = remote_data.get('private_url')
        if not token or not api_address or not private_api_address:
            # The Gateway hasn't published info yet
            return
        self.state.api_address = api_address.replace('"', '')
        self.state.private_api_address = private_api_address.replace('"', '')
        self.state.client_token = token.replace('"', '')
        self.on.gateway_changed.emit()

    def on_relation_departed(self, event):
        self.state.api_address = None
        self.state.client_token = None
        self.state.allow_insecure = None
        self.on.gateway_changed.emit()

    @property
    def is_joined(self) -> bool:
        return self.framework.model.get_relation(self._relation_name) is not None

    @property
    def is_ready(self) -> bool:
        return bool(self.state.api_address and self.client_token)

    @property
    def client_token(self) -> str:
        return self.state.client_token

    @property
    def private_api_address(self) -> str:
        return self.state.private_api_address

    @property
    def api_address(self) -> str:
        return self.state.api_address
