#!/usr/bin/env python3
# Anbox Cloud Dashboard
# (c) 2021 Canonical Ltd. All right reservered
#

import logging

from ops.charm import CharmBase, CharmEvents, RelationChangedEvent
from ops.framework import Object, EventBase, EventSource, StoredState

logger = logging.getLogger(__name__)


class HttpChangedEvent(EventBase):
    pass


class HttpDepartedEvent(EventBase):
    pass


class HttpEvents(CharmEvents):
    http_changed = EventSource(HttpChangedEvent)
    http_departed = EventSource(HttpDepartedEvent)


class Http(Object):
    on = HttpEvents()
    state = StoredState()

    def __init__(self, charm: CharmBase, relation_name: str):
        super().__init__(charm, relation_name)

        self._relation_name = relation_name

        self.state.set_default(remote_public_address=None)

        self.framework.observe(charm.on[relation_name].relation_changed, self.on_relation_changed)
        self.framework.observe(charm.on[relation_name].relation_departed, self.on_relation_departed)

    def on_relation_changed(self, event: RelationChangedEvent):
        remote_data = event.relation.data[next(iter(event.relation.units))]
        remote_public_address = remote_data.get('public-address')
        if not remote_public_address:
            return
        self.state.remote_public_address = remote_public_address.replace('"', '')
        self.on.http_changed.emit()


    def on_relation_departed(self, event):
        self.state.api_address = None
        self.state.client_token = None
        self.state.allow_insecure = None
        self.state.remote_public_address = None
        self.on.http_departed.emit()


    def configure(self, port, service_name, hostname=None, ):
        """
        Send backend information to HAProxy on the http relation.
        service_name MUST be defined in the HAProxy configuration under
        'services' to work.
        """
        for unit in self.model.get_relation(self._relation_name).units:
            if unit == self.model.unit:
                continue

            relation = self.model.get_relation(self._relation_name)
            local_data = relation.data[self.model.unit]

            ingress_address = local_data.get('ingress-address') or local_data.get('private-address')

            local_data['service_name'] = service_name
            local_data['hostname'] = hostname or ingress_address
            local_data['port'] = str(port)

    @property
    def remote_public_address(self) -> str:
        return self.state.remote_public_address